==> fred-akm-ng_1.1.4.orig.tar.gz: fred-akm-ng-1.1.4/src/database/process_acceptable_keysets.cc <==
/*
 * Copyright (C) 2021-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/database/impl/add_keyset.hh"
#include "src/database/impl/extended_info_domain_data.hh"
#include "src/database/impl/is_domain_ok_for_update.hh"
#include "src/database/impl/is_keyset_ok_for_update.hh"
#include "src/database/impl/turn_off_dnssec.hh"
#include "src/database/impl/update_keyset.hh"
#include "src/database/process_acceptable_keysets.hh"
#include "src/exception.hh"
#include "src/scan_type.hh"

#include <boost/variant.hpp>

#include <libgrill/exception.hh>

#include <libfred/opcontext.hh>
#include <libfred/registrable_object/domain/info_domain.hh>

#include <liblog/liblog.hh>

#include <algorithm>
#include <cstdint>

namespace Fred {
namespace Akm {
namespace Database {

namespace {

enum class Success
{
    no,
    yes
};

struct UpdateStatus
{
    struct DomainNotFound {};

    struct NotReadyToUpdate
    {
        ScanType scan_type;
    };

    struct ReadyToUpdate
    {
        ScanType scan_type;
        Success result;
    };

    using Type = boost::variant<DomainNotFound, NotReadyToUpdate, ReadyToUpdate>;
};

void akm_turn_off(
        const LibFred::OperationContextCreator& ctx,
        const LibFred::InfoDomainData& domain_info,
        const KeysetMetadata& keyset_metadata,
        std::function<std::unique_ptr<Logger::LogEntry<Logger::LogEntryType::akm_turn_off>>()> make_log_entry)
{
    LIBLOG_SET_CONTEXT(LogCtx, log_ctx, __func__, domain_info.fqdn);
    Impl::turn_off_dnssec(ctx, domain_info, keyset_metadata, make_log_entry);
}

void akm_rollover(
        const LibFred::OperationContextCreator& ctx,
        const Keyset& keyset,
        const LibFred::InfoDomainData& domain_info,
        const LibFred::InfoKeysetData& keyset_info,
        const KeysetMetadata& keyset_metadata,
        std::function<std::unique_ptr<Logger::LogEntry<Logger::LogEntryType::akm_rollover>>()> make_log_entry)
{
    LIBLOG_SET_CONTEXT(LogCtx, log_ctx, __func__, domain_info.fqdn);
    Impl::update_keyset(ctx, domain_info, keyset_info, keyset_metadata, keyset, make_log_entry);
}

void akm_turn_on(
        const LibFred::OperationContextCreator& ctx,
        const Keyset& keyset,
        const LibFred::InfoDomainData& domain_info,
        const KeysetMetadata& keyset_metadata,
        std::function<std::unique_ptr<Logger::LogEntry<Logger::LogEntryType::akm_turn_on>>()> make_log_entry)
{
    LIBLOG_SET_CONTEXT(LogCtx, log_ctx, __func__, domain_info.fqdn);
    Impl::add_keyset(ctx, domain_info, keyset_metadata, keyset, make_log_entry);
}


bool is_domain_changed(
        const LibFred::OperationContext& ctx,
        const DomainUuid &_domain_uuid)
{
    ::Database::Result r = ctx.get_conn().exec_params(
        // clang-format off
        "SELECT "
          "FROM domain_history p_dh "
          "JOIN history h ON h.id = p_dh.historyid "
          "JOIN domain_history n_dh ON n_dh.historyid = h.next "
          "JOIN object_registry oreg ON oreg.id = n_dh.id "
         "WHERE ((p_dh.nsset IS DISTINCT FROM n_dh.nsset) OR (p_dh.keyset IS DISTINCT FROM n_dh.keyset)) "
           "AND NOW() - '3 DAY'::INTERVAL < h.valid_to "
           "AND oreg.uuid = $1::UUID",
        // clang-format on
        ::Database::query_param_list(*_domain_uuid));

    return r.size() > 0;
}

bool is_nsset_changed(
        const LibFred::OperationContext& ctx,
        unsigned long long nsset_id)
{
    ::Database::Result r = ctx.get_conn().exec_params(
        // clang-format off
        "SELECT "
          "FROM object_registry oreg "
          "JOIN object o ON o.id = oreg.id "
         "WHERE oreg.id = $1::BIGINT "
           "AND oreg.type = get_object_type_id('nsset') "
           "AND NOW() - '3 DAY'::INTERVAL < o.update",
        // clang-format on
        ::Database::query_param_list(nsset_id));

    return r.size() > 0;
}

bool is_keyset_changed(
        const LibFred::OperationContext& ctx,
        unsigned long long keyset_id)
{
    ::Database::Result r = ctx.get_conn().exec_params(
        // clang-format off
        "SELECT "
          "FROM object_registry oreg "
          "JOIN object o ON o.id = oreg.id "
         "WHERE oreg.id = $1::BIGINT "
           "AND oreg.type = get_object_type_id('keyset') "
           "AND NOW() - '3 DAY'::INTERVAL < o.update",
        // clang-format on
        ::Database::query_param_list(keyset_id));

    return r.size() > 0;
}


UpdateStatus::Type process_one_domain(
        const KeysetMetadata& _keyset_metadata,
        const AcceptableKeyset& _acceptable_keyset,
        Logger::LogEntryFactory& _log_entry_factory,
        DryRun dry_run)
{
    LIBLOG_SET_CONTEXT(LogCtx, log_ctx, __func__, *_acceptable_keyset.second.domain_fqdn);
    boost::optional<ExtendedInfoDomainData> extended_domain_info;
    try
    {
        LibFred::OperationContextCreator ctx;
        LIBLOG_DEBUG("processing keyset for domain {}: uuid {}, keyset size {}",
                *_acceptable_keyset.second.domain_fqdn,
                to_string(domain_uuid(_acceptable_keyset)),
                _acceptable_keyset.second.keyset.size());
        std::for_each(
                cbegin(_acceptable_keyset.second.keyset),
                cend(_acceptable_keyset.second.keyset),
                [fqdn = *_acceptable_keyset.second.domain_fqdn](auto&& dnskey)
        {
            LIBLOG_DEBUG("domain {}: {}, {}, {}, {}", fqdn, *dnskey.flags, *dnskey.proto, *dnskey.alg, *dnskey.public_key);
        });

        extended_domain_info = get_domain_info_from_fqdn(
                ctx,
                _keyset_metadata,
                _acceptable_keyset.second.domain_fqdn);
        if (extended_domain_info == boost::none)
        {
            LIBLOG_INFO("domain not found");
            return UpdateStatus::DomainNotFound{};
        }

        // XXX: quickfix
        LIBLOG_INFO("check for domain and linked nsset and linked keyset changes");
        auto has_nsset = !extended_domain_info->info_domain_data.nsset.isnull();
        auto has_keyset = !extended_domain_info->info_domain_data.keyset.isnull();
        if (is_domain_changed(ctx, domain_uuid(_acceptable_keyset))
            || (has_nsset && is_nsset_changed(ctx, extended_domain_info->info_domain_data.nsset.get_value().id))
            || (has_keyset && is_keyset_changed(ctx, extended_domain_info->info_domain_data.keyset.get_value().id)))
        {
            LIBLOG_INFO("domain or linked nsset or linked keyset was changed");
            return UpdateStatus::NotReadyToUpdate{extended_domain_info->scan_type};
        }
        LIBLOG_INFO("domain and linked nsset and linked keyset was unchanged");
        // XXX: quickfix

        const auto domain_is_ok_for_update = Impl::is_domain_ok_for_update(
                ctx,
                extended_domain_info->info_domain_data,
                domain_uuid(_acceptable_keyset),
                _keyset_metadata);
        if (!domain_is_ok_for_update)
        {
            LIBLOG_INFO("domain is not OK for update");
            return UpdateStatus::NotReadyToUpdate{extended_domain_info->scan_type};
        }
        const auto make_successful_result = [&]()
        {
            if (dry_run != DryRun::yes)
            {
                ctx.commit_transaction();
            }
            return UpdateStatus::ReadyToUpdate{extended_domain_info->scan_type, Success::yes};
        };
        if (Impl::is_dnssec_turn_off_requested(_acceptable_keyset.second.keyset))
        {
            if (extended_domain_info->scan_type == ScanType::insecure)
            {
                LIBLOG_INFO("turn off requested for insecure domain, nothing to do.");
                return UpdateStatus::NotReadyToUpdate{extended_domain_info->scan_type};
            }
            const auto create_log_entry = [&]() -> std::unique_ptr<Logger::LogEntry<Logger::LogEntryType::akm_turn_off>>
            {
                switch (dry_run)
                {
                    case DryRun::no:
                        return _log_entry_factory.create_log_entry<Logger::LogEntryType::akm_turn_off>(
                                DomainId{extended_domain_info->info_domain_data.id},
                                _acceptable_keyset.second.domain_fqdn);
                    case DryRun::yes:
                        return nullptr;
                }
                throw std::runtime_error{"unexpected DryRun value"};
            };
            akm_turn_off(
                    ctx,
                    extended_domain_info->info_domain_data,
                    _keyset_metadata,
                    create_log_entry);
            return make_successful_result();
        }
        if (!Impl::is_keyset_ok_for_update(ctx, _acceptable_keyset.second.keyset))
        {
            LIBLOG_INFO("keyset is not OK for update");
            return UpdateStatus::NotReadyToUpdate{extended_domain_info->scan_type};
        }

        if (extended_domain_info->scan_type == ScanType::secure_auto)
        {
            const auto create_log_entry = [&]() -> std::unique_ptr<Logger::LogEntry<Logger::LogEntryType::akm_rollover>>
            {
                switch (dry_run)
                {
                    case DryRun::no:
                        return _log_entry_factory.create_log_entry<Logger::LogEntryType::akm_rollover>(
                                DomainId{extended_domain_info->info_domain_data.id},
                                _acceptable_keyset.second.domain_fqdn);
                    case DryRun::yes:
                        return nullptr;
                }
                throw std::runtime_error{"unexpected DryRun value"};
            };
            akm_rollover(
                    ctx,
                    _acceptable_keyset.second.keyset,
                    extended_domain_info->info_domain_data,
                    *extended_domain_info->info_keyset_data,
                    _keyset_metadata,
                    create_log_entry);
            return make_successful_result();
        }
        const auto create_log_entry = [&]() -> std::unique_ptr<Logger::LogEntry<Logger::LogEntryType::akm_turn_on>>
        {
            switch (dry_run)
            {
                case DryRun::no:
                    return _log_entry_factory.create_log_entry<Logger::LogEntryType::akm_turn_on>(
                            DomainId{extended_domain_info->info_domain_data.id},
                            _acceptable_keyset.second.domain_fqdn);
                case DryRun::yes:
                    return nullptr;
            }
            throw std::runtime_error{"unexpected DryRun value"};
        };
        akm_turn_on(
                ctx,
                _acceptable_keyset.second.keyset,
                extended_domain_info->info_domain_data,
                _keyset_metadata,
                create_log_entry);
        return make_successful_result();
    }
    catch (const LibGrill::GrpcException& e) // cannot be caught by std::exception; TODO: fix libgrill
    {
        LIBLOG_ERROR("LibGrill::GrpcException caught: {}", e.error_message());
        throw ConfigurationError{"logger misconfigured"};
    }
    catch (const ConfigurationError& e)
    {
        LIBLOG_ERROR("ConfigurationError exception: {}", e.what());
        throw;
    }
    catch (const Impl::UpdateKeysetFailed& e)
    {
        LIBLOG_INFO("UpdateKeysetFailed exception: {}", e.what());
    }
    catch (const Impl::OperationFailed& e)
    {
        LIBLOG_INFO("OperationFailed exception: {}", e.what());
    }
    catch (const std::exception& e)
    {
        LIBLOG_INFO("std::exception: {}", e.what());
    }
    catch (...)
    {
        LIBLOG_ERROR("unknown error occurred");
    }
    if (extended_domain_info != boost::none)
    {
        return UpdateStatus::ReadyToUpdate{extended_domain_info->scan_type, Success::no};
    }
    return UpdateStatus::DomainNotFound{};
}

class Statistics
{
public:
    void add(const UpdateStatus::Type& update_status)
    {
        boost::apply_visitor(*this, update_status);
    }
    void print()
    {
        std::for_each(cbegin(type_to_data_), cend(type_to_data_), [](auto&& type_data_pair)
        {
            const auto& type = type_data_pair.first;
            const auto& data = type_data_pair.second;
            LIBLOG_INFO("======== update stats {} candidates =======", to_string(type));
            LIBLOG_INFO("domains loaded:         {}", data.total());
            LIBLOG_INFO("|- ok_for_update:       {}", data.ok_for_update());
            LIBLOG_INFO("|   |- updated:         {}", data.ok_for_update_success);
            LIBLOG_INFO("|   -- failed:          {}", data.ok_for_update_failure);
            LIBLOG_INFO("-- ko_for_update:       {}", data.not_ok_for_update);
        });
        if (domains_not_found_ != 0)
        {
            LIBLOG_WARNING("not found domains:       {}", domains_not_found_);
        }
    }
    void operator()(const UpdateStatus::DomainNotFound&)
    {
        ++domains_not_found_;
    }
    void operator()(const UpdateStatus::NotReadyToUpdate& status)
    {
        type_to_data_[status.scan_type].add_not_ready();
    }
    void operator()(const UpdateStatus::ReadyToUpdate& status)
    {
        type_to_data_[status.scan_type].add_ready(status.result);
    }
private:
    struct Data
    {
        std::int64_t ok_for_update_success = 0;
        std::int64_t ok_for_update_failure = 0;
        std::int64_t not_ok_for_update = 0;
        constexpr std::int64_t ok_for_update() const noexcept
        {
            return ok_for_update_success + ok_for_update_failure;
        }
        constexpr std::int64_t total() const noexcept
        {
            return ok_for_update() + not_ok_for_update;
        }
        constexpr void add_not_ready()
        {
            ++not_ok_for_update;
        }
        constexpr void add_ready(Success result)
        {
            switch (result)
            {
                case Success::no:
                    ++ok_for_update_failure;
                    return;
                case Success::yes:
                    ++ok_for_update_success;
                    return;
            }
        }
    };
    std::int64_t domains_not_found_ = 0;
    std::map<ScanType, Data> type_to_data_ = {
            {ScanType::insecure, {}},
            {ScanType::secure, {}},
            {ScanType::secure_auto, {}}};
};

} // namespace Fred::Akm::Database::{anonymous}

void process_acceptable_keysets(
        const KeysetMetadata& _keyset_metadata,
        const AcceptableKeysets& _acceptable_keysets,
        Logger::LogEntryFactory& _log_entry_factory,
        DryRun dry_run)
{
    Statistics statistics;
    std::for_each(cbegin(_acceptable_keysets), cend(_acceptable_keysets), [&](auto&& acceptable_keyset)
    {
        const auto update_status = process_one_domain(
                _keyset_metadata,
                acceptable_keyset,
                _log_entry_factory,
                dry_run);
        statistics.add(update_status);
    });
    statistics.print();
}

} // namespace Fred::Akm::Database
} // namespace Fred::Akm
} // namespace Fred

==> fred-akm-ng_1.1.5.orig.tar.gz: fred-akm-ng-1.1.5/src/database/process_acceptable_keysets.cc <==
/*
 * Copyright (C) 2021-2025  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/database/impl/add_keyset.hh"
#include "src/database/impl/extended_info_domain_data.hh"
#include "src/database/impl/is_domain_ok_for_update.hh"
#include "src/database/impl/is_keyset_ok_for_update.hh"
#include "src/database/impl/turn_off_dnssec.hh"
#include "src/database/impl/update_keyset.hh"
#include "src/database/process_acceptable_keysets.hh"
#include "src/exception.hh"
#include "src/scan_type.hh"

#include <boost/variant.hpp>

#include <libgrill/exception.hh>

#include <libfred/opcontext.hh>
#include <libfred/registrable_object/domain/info_domain.hh>

#include <liblog/liblog.hh>

#include <algorithm>
#include <cstdint>
#include <iomanip>
#include <sstream>

namespace Fred {
namespace Akm {
namespace Database {

namespace {

enum class Success
{
    no,
    yes
};

struct UpdateStatus
{
    struct DomainNotFound {};

    struct NotReadyToUpdate
    {
        ScanType scan_type;
    };

    struct ReadyToUpdate
    {
        ScanType scan_type;
        Success result;
    };

    using Type = boost::variant<DomainNotFound, NotReadyToUpdate, ReadyToUpdate>;
};

void akm_turn_off(
        const LibFred::OperationContextCreator& ctx,
        const LibFred::InfoDomainData& domain_info,
        const KeysetMetadata& keyset_metadata,
        std::function<std::unique_ptr<Logger::LogEntry<Logger::LogEntryType::akm_turn_off>>()> make_log_entry)
{
    LIBLOG_SET_CONTEXT(LogCtx, log_ctx, __func__, domain_info.fqdn);
    Impl::turn_off_dnssec(ctx, domain_info, keyset_metadata, make_log_entry);
}

void akm_rollover(
        const LibFred::OperationContextCreator& ctx,
        const Keyset& keyset,
        const LibFred::InfoDomainData& domain_info,
        const LibFred::InfoKeysetData& keyset_info,
        const KeysetMetadata& keyset_metadata,
        std::function<std::unique_ptr<Logger::LogEntry<Logger::LogEntryType::akm_rollover>>()> make_log_entry)
{
    LIBLOG_SET_CONTEXT(LogCtx, log_ctx, __func__, domain_info.fqdn);
    Impl::update_keyset(ctx, domain_info, keyset_info, keyset_metadata, keyset, make_log_entry);
}

void akm_turn_on(
        const LibFred::OperationContextCreator& ctx,
        const Keyset& keyset,
        const LibFred::InfoDomainData& domain_info,
        const KeysetMetadata& keyset_metadata,
        std::function<std::unique_ptr<Logger::LogEntry<Logger::LogEntryType::akm_turn_on>>()> make_log_entry)
{
    LIBLOG_SET_CONTEXT(LogCtx, log_ctx, __func__, domain_info.fqdn);
    Impl::add_keyset(ctx, domain_info, keyset_metadata, keyset, make_log_entry);
}

std::string to_seconds_string(const std::chrono::microseconds& t)
{
    std::ostringstream out;
    out << (t.count() / 1'000'000) << "." << std::setw(6) << std::setfill('0') << (t.count() % 1'000'000);
    return out.str();
}

bool is_domain_changed(
        const LibFred::OperationContext& ctx,
        const DomainUuid &_domain_uuid,
        const std::chrono::microseconds& unchanged_from)
{
    ::Database::Result r = ctx.get_conn().exec_params(
        // clang-format off
        "SELECT "
          "FROM domain_history p_dh "
          "JOIN history h ON h.id = p_dh.historyid "
          "JOIN domain_history n_dh ON n_dh.historyid = h.next "
          "JOIN object_registry oreg ON oreg.id = n_dh.id "
         "WHERE ((p_dh.nsset IS DISTINCT FROM n_dh.nsset) OR (p_dh.keyset IS DISTINCT FROM n_dh.keyset)) "
           "AND TO_TIMESTAMP($2::DOUBLE PRECISION) < h.valid_to "
           "AND oreg.uuid = $1::UUID "
         "LIMIT 1",
        // clang-format on
        ::Database::QueryParams{*_domain_uuid, to_seconds_string(unchanged_from)});

    return r.size() != 0;
}

bool is_nsset_changed(
        const LibFred::OperationContext& ctx,
        unsigned long long nsset_id,
        const std::chrono::microseconds& unchanged_from)
{
    ::Database::Result r = ctx.get_conn().exec_params(
        // clang-format off
        "SELECT "
          "FROM object_registry oreg "
          "JOIN object o ON o.id = oreg.id "
         "WHERE oreg.id = $1::BIGINT "
           "AND oreg.type = get_object_type_id('nsset') "
           "AND TO_TIMESTAMP($2::DOUBLE PRECISION) < o.update "
         "LIMIT 1",
        // clang-format on
        ::Database::QueryParams{nsset_id, to_seconds_string(unchanged_from)});

    return r.size() != 0;
}

bool is_keyset_changed(
        const LibFred::OperationContext& ctx,
        unsigned long long keyset_id,
        const std::chrono::microseconds& unchanged_from)
{
    ::Database::Result r = ctx.get_conn().exec_params(
        // clang-format off
        "SELECT "
          "FROM object_registry oreg "
          "JOIN object o ON o.id = oreg.id "
         "WHERE oreg.id = $1::BIGINT "
           "AND oreg.type = get_object_type_id('keyset') "
           "AND TO_TIMESTAMP($2::DOUBLE PRECISION) < o.update "
         "LIMIT 1",
        // clang-format on
        ::Database::QueryParams{keyset_id, to_seconds_string(unchanged_from)});

    return r.size() != 0;
}


UpdateStatus::Type process_one_domain(
        const KeysetMetadata& _keyset_metadata,
        const AcceptableKeyset& _acceptable_keyset,
        Logger::LogEntryFactory& _log_entry_factory,
        DryRun dry_run)
{
    LIBLOG_SET_CONTEXT(LogCtx, log_ctx, __func__, *_acceptable_keyset.second.domain_fqdn);
    boost::optional<ExtendedInfoDomainData> extended_domain_info;
    try
    {
        LibFred::OperationContextCreator ctx;
        LIBLOG_DEBUG("processing keyset for domain {}: uuid {}, keyset size {}",
                *_acceptable_keyset.second.domain_fqdn,
                to_string(domain_uuid(_acceptable_keyset)),
                _acceptable_keyset.second.keyset.size());
        std::for_each(
                cbegin(_acceptable_keyset.second.keyset),
                cend(_acceptable_keyset.second.keyset),
                [fqdn = *_acceptable_keyset.second.domain_fqdn](auto&& dnskey)
        {
            LIBLOG_DEBUG("domain {}: {}, {}, {}, {}", fqdn, *dnskey.flags, *dnskey.proto, *dnskey.alg, *dnskey.public_key);
        });

        extended_domain_info = get_domain_info_from_fqdn(
                ctx,
                _keyset_metadata,
                _acceptable_keyset.second.domain_fqdn);
        if (extended_domain_info == boost::none)
        {
            LIBLOG_INFO("domain not found");
            return UpdateStatus::DomainNotFound{};
        }

        LIBLOG_DEBUG("check for domain and linked nsset and linked keyset changes");
        const auto has_nsset = [&]() { return !extended_domain_info->info_domain_data.nsset.isnull(); };
        const auto has_keyset = [&]() { return !extended_domain_info->info_domain_data.keyset.isnull(); };
        static constexpr auto protection_interval = std::chrono::hours{1};
        const auto unchanged_from = _acceptable_keyset.second.imported_at - protection_interval;
        if (is_domain_changed(ctx, domain_uuid(_acceptable_keyset), unchanged_from) ||
            (has_nsset() && is_nsset_changed(ctx, extended_domain_info->info_domain_data.nsset.get_value().id, unchanged_from)) ||
            (has_keyset() && is_keyset_changed(ctx, extended_domain_info->info_domain_data.keyset.get_value().id, unchanged_from)))
        {
            LIBLOG_INFO("domain or linked nsset or linked keyset was changed");
            return UpdateStatus::NotReadyToUpdate{extended_domain_info->scan_type};
        }
        LIBLOG_DEBUG("domain and linked nsset and linked keyset was unchanged");

        const auto domain_is_ok_for_update = Impl::is_domain_ok_for_update(
                ctx,
                extended_domain_info->info_domain_data,
                domain_uuid(_acceptable_keyset),
                _keyset_metadata);
        if (!domain_is_ok_for_update)
        {
            LIBLOG_INFO("domain is not OK for update");
            return UpdateStatus::NotReadyToUpdate{extended_domain_info->scan_type};
        }
        const auto make_successful_result = [&]()
        {
            if (dry_run != DryRun::yes)
            {
                ctx.commit_transaction();
            }
            return UpdateStatus::ReadyToUpdate{extended_domain_info->scan_type, Success::yes};
        };
        if (Impl::is_dnssec_turn_off_requested(_acceptable_keyset.second.keyset))
        {
            if (extended_domain_info->scan_type == ScanType::insecure)
            {
                LIBLOG_INFO("turn off requested for insecure domain, nothing to do.");
                return UpdateStatus::NotReadyToUpdate{extended_domain_info->scan_type};
            }
            const auto create_log_entry = [&]() -> std::unique_ptr<Logger::LogEntry<Logger::LogEntryType::akm_turn_off>>
            {
                switch (dry_run)
                {
                    case DryRun::no:
                        return _log_entry_factory.create_log_entry<Logger::LogEntryType::akm_turn_off>(
                                DomainId{extended_domain_info->info_domain_data.id},
                                _acceptable_keyset.second.domain_fqdn);
                    case DryRun::yes:
                        return nullptr;
                }
                throw std::runtime_error{"unexpected DryRun value"};
            };
            akm_turn_off(
                    ctx,
                    extended_domain_info->info_domain_data,
                    _keyset_metadata,
                    create_log_entry);
            return make_successful_result();
        }
        if (!Impl::is_keyset_ok_for_update(ctx, _acceptable_keyset.second.keyset))
        {
            LIBLOG_INFO("keyset is not OK for update");
            return UpdateStatus::NotReadyToUpdate{extended_domain_info->scan_type};
        }

        if (extended_domain_info->scan_type == ScanType::secure_auto)
        {
            const auto create_log_entry = [&]() -> std::unique_ptr<Logger::LogEntry<Logger::LogEntryType::akm_rollover>>
            {
                switch (dry_run)
                {
                    case DryRun::no:
                        return _log_entry_factory.create_log_entry<Logger::LogEntryType::akm_rollover>(
                                DomainId{extended_domain_info->info_domain_data.id},
                                _acceptable_keyset.second.domain_fqdn);
                    case DryRun::yes:
                        return nullptr;
                }
                throw std::runtime_error{"unexpected DryRun value"};
            };
            akm_rollover(
                    ctx,
                    _acceptable_keyset.second.keyset,
                    extended_domain_info->info_domain_data,
                    *extended_domain_info->info_keyset_data,
                    _keyset_metadata,
                    create_log_entry);
            return make_successful_result();
        }
        const auto create_log_entry = [&]() -> std::unique_ptr<Logger::LogEntry<Logger::LogEntryType::akm_turn_on>>
        {
            switch (dry_run)
            {
                case DryRun::no:
                    return _log_entry_factory.create_log_entry<Logger::LogEntryType::akm_turn_on>(
                            DomainId{extended_domain_info->info_domain_data.id},
                            _acceptable_keyset.second.domain_fqdn);
                case DryRun::yes:
                    return nullptr;
            }
            throw std::runtime_error{"unexpected DryRun value"};
        };
        akm_turn_on(
                ctx,
                _acceptable_keyset.second.keyset,
                extended_domain_info->info_domain_data,
                _keyset_metadata,
                create_log_entry);
        return make_successful_result();
    }
    catch (const LibGrill::GrpcException& e) // cannot be caught by std::exception; TODO: fix libgrill
    {
        LIBLOG_ERROR("LibGrill::GrpcException caught: {}", e.error_message());
        throw ConfigurationError{"logger misconfigured"};
    }
    catch (const ConfigurationError& e)
    {
        LIBLOG_ERROR("ConfigurationError exception: {}", e.what());
        throw;
    }
    catch (const Impl::UpdateKeysetFailed& e)
    {
        LIBLOG_INFO("UpdateKeysetFailed exception: {}", e.what());
    }
    catch (const Impl::OperationFailed& e)
    {
        LIBLOG_INFO("OperationFailed exception: {}", e.what());
    }
    catch (const std::exception& e)
    {
        LIBLOG_INFO("std::exception: {}", e.what());
    }
    catch (...)
    {
        LIBLOG_ERROR("unknown error occurred");
    }
    if (extended_domain_info != boost::none)
    {
        return UpdateStatus::ReadyToUpdate{extended_domain_info->scan_type, Success::no};
    }
    return UpdateStatus::DomainNotFound{};
}

class Statistics
{
public:
    void add(const UpdateStatus::Type& update_status)
    {
        boost::apply_visitor(*this, update_status);
    }
    void print()
    {
        std::for_each(cbegin(type_to_data_), cend(type_to_data_), [](auto&& type_data_pair)
        {
            const auto& type = type_data_pair.first;
            const auto& data = type_data_pair.second;
            LIBLOG_INFO("======== update stats {} candidates =======", to_string(type));
            LIBLOG_INFO("domains loaded:         {}", data.total());
            LIBLOG_INFO("|- ok_for_update:       {}", data.ok_for_update());
            LIBLOG_INFO("|   |- updated:         {}", data.ok_for_update_success);
            LIBLOG_INFO("|   -- failed:          {}", data.ok_for_update_failure);
            LIBLOG_INFO("-- ko_for_update:       {}", data.not_ok_for_update);
        });
        if (domains_not_found_ != 0)
        {
            LIBLOG_WARNING("not found domains:       {}", domains_not_found_);
        }
    }
    void operator()(const UpdateStatus::DomainNotFound&)
    {
        ++domains_not_found_;
    }
    void operator()(const UpdateStatus::NotReadyToUpdate& status)
    {
        type_to_data_[status.scan_type].add_not_ready();
    }
    void operator()(const UpdateStatus::ReadyToUpdate& status)
    {
        type_to_data_[status.scan_type].add_ready(status.result);
    }
private:
    struct Data
    {
        std::int64_t ok_for_update_success = 0;
        std::int64_t ok_for_update_failure = 0;
        std::int64_t not_ok_for_update = 0;
        constexpr std::int64_t ok_for_update() const noexcept
        {
            return ok_for_update_success + ok_for_update_failure;
        }
        constexpr std::int64_t total() const noexcept
        {
            return ok_for_update() + not_ok_for_update;
        }
        constexpr void add_not_ready()
        {
            ++not_ok_for_update;
        }
        constexpr void add_ready(Success result)
        {
            switch (result)
            {
                case Success::no:
                    ++ok_for_update_failure;
                    return;
                case Success::yes:
                    ++ok_for_update_success;
                    return;
            }
        }
    };
    std::int64_t domains_not_found_ = 0;
    std::map<ScanType, Data> type_to_data_ = {
            {ScanType::insecure, {}},
            {ScanType::secure, {}},
            {ScanType::secure_auto, {}}};
};

} // namespace Fred::Akm::Database::{anonymous}

void process_acceptable_keysets(
        const KeysetMetadata& _keyset_metadata,
        const AcceptableKeysets& _acceptable_keysets,
        Logger::LogEntryFactory& _log_entry_factory,
        DryRun dry_run)
{
    Statistics statistics;
    std::for_each(cbegin(_acceptable_keysets), cend(_acceptable_keysets), [&](auto&& acceptable_keyset)
    {
        const auto update_status = process_one_domain(
                _keyset_metadata,
                acceptable_keyset,
                _log_entry_factory,
                dry_run);
        statistics.add(update_status);
    });
    statistics.print();
}

} // namespace Fred::Akm::Database
} // namespace Fred::Akm
} // namespace Fred
